/*
Universidad Nacional de Lujn
Departamento de Ciencias Bsicas
Division Estadstica y Sistemas
Base de Datos
Lic. Guillermo Cherencio

Implementacion de Entidad Fuerte - Debil con FireBird / InterBase

El discriminador de la tabla debil debe ser un numero correlativo
ascendente por cada valor de p.k. de la entidad fuerte
Dadas las siguientes relaciones:
	fuerte(clave,descr) 
		clave p.k.
	debil(clave,discr)  
		clave,discr p.k.
		clave f.k -> fuerte
Una posible implementacion seria:
	Crear una tabla paralela a fuerte, llamada fuerte2 de la 
siguiente manera:
	fuerte2(clave,ultimo)
		clave p.k.
	el atributo ultimo representa el ultimo valor de "discr"
relacionado con dicha "clave".  Toda la actualizacion y consulta
sobre la tabla paralela fuerte2 se realiza a traves de stores pro-
cedures, de modo tal, que el usuario podra tener permiso de ejecu-
cion sobre los stores procedures que la actualizan o consultan, pe-
ro no tendra permiso sobre la tabla fuerte2, a lo sumo tendra permi-
so SELECT.  Esto nos evita que un usuario pueda actualizarnos deli-
beradamente la tabla fuerte2 y rompa el control automatico que inten-
tamos realizar.  No obstante, es posible que el usuario intente eje-
cutar aisladamente los store procedures, para lo cual debera saber
el nombre de los mismos y pasarle los parametros validos, de manera
correcta, para realizar su ejecucion. Un control sencillo seria agregar
un parametro a los store procedures que sea enviado desde los triggers
y dentro de los store procedures controlar el contenido del mismo.
Seria muy util que el SGBD contara con la opcion de encriptacion para
mantener el codigo de estos store procedures encriptados.
	La idea es que toda la actualizacion de la tabla fuerte2 se 
realice a traves de los store procedures y dichos store procedures sean
ejecutados automaticamente a traves del control de los eventos de actua-
lizacion, insercin y borrado de las tablas fuerte y debil.

Modo de Ejecucion:  	
   isql ferreteria.gdb -u SYSDBA -p masterkey -i dersql1.sql -o salida.txt

*/
SET ECHO ON;
SET;

DROP TRIGGER trg_aifuerte;      /* evento after insert tabla fuerte       */
DROP TRIGGER trg_adfuerte;      /* evento after delete tabla fuerte       */
DROP TRIGGER trg_aufuerte;      /* evento after update tabla fuerte       */
DROP TRIGGER trg_bidebil;       /* evento before insert tabla debil       */
DROP TRIGGER trg_aidebil;       /* evento after  insert tabla debil       */
DROP TRIGGER trg_budebil;       /* evento before update tabla debil       */
DROP TRIGGER trg_bddebil;       /* evento before delete tabla debil       */
DROP TRIGGER trg_addebil;       /* evento after  delete tabla debil       */
DROP EXCEPTION ex_discr_nocorr; /* Error de correlatividad                */
DROP EXCEPTION ex_discr_noupd;  /* Error update en tabla debil            */
DROP EXCEPTION ex_discr_nodel;  /* Error delete en tabla debil            */
DROP PROCEDURE sp_idfuerte2;    /* sp para insert,delete de tabla fuerte2 */
DROP PROCEDURE sp_ucfuerte2;    /* sp para update claves de tabla fuerte2 */
DROP PROCEDURE sp_ufuerte2;     /* sp para update        de tabla fuerte2 */
DROP PROCEDURE sp_sfuerte2;     /* sp para select        de tabla fuerte2 */
DROP TABLE debil;
DROP TABLE fuerte;
DROP TABLE fuerte2;

/* Creo Tabla paralela fuerte2 */
CREATE TABLE fuerte2(  /* Mantenida automaticamente por Store Procedures */
   clave NUMERIC(3,0) DEFAULT 1 NOT NULL,
   ultimo NUMERIC(2,0) NOT NULL, /* ultimo valor de discriminador */
   PRIMARY KEY (clave)
);

/* Creo Tabla fuerte */
CREATE TABLE fuerte(
   clave NUMERIC(3,0) DEFAULT 1 NOT NULL,
   descr VARCHAR(40) NOT NULL,
   PRIMARY KEY (clave)
);

/* Creo Tabla debil */
CREATE TABLE debil(
   clave NUMERIC(3,0) DEFAULT 1 NOT NULL REFERENCES fuerte,  /* pk entidad fuerte */
   discr NUMERIC(2,0) DEFAULT 1 NOT NULL,  /* discriminador */
   PRIMARY KEY (clave,discr)
);

/* Creo Procedimiento para insert,delete en tabla fuerte2 */
SET TERM !! ;
CREATE PROCEDURE sp_idfuerte2 (operacion char(1),idclave numeric(3,0))
AS
BEGIN
   IF ( :operacion = 'I' ) THEN INSERT INTO fuerte2 VALUES(:idclave,0);
   IF ( :operacion = 'D' ) THEN DELETE FROM fuerte2 WHERE clave = :idclave; 
END !!
SET TERM ; !!

/* Creo Procedimiento para cambio de claves en tabla fuerte2 */
SET TERM !! ;
CREATE PROCEDURE sp_ucfuerte2 (idclaveold numeric(3,0),idclavenew numeric(3,0))
AS
BEGIN
   UPDATE fuerte2 SET clave = :idclavenew WHERE clave = :idclaveold;
END !!
SET TERM ; !!

/* Creo Procedimiento para incrementar/decrementar contador en fuerte2 */
SET TERM !! ;
CREATE PROCEDURE sp_ufuerte2 (idclave numeric(3,0),incremento numeric(2))
AS
BEGIN
   UPDATE fuerte2 SET ultimo = ultimo + :incremento WHERE clave = :idclave;
END !!
SET TERM ; !!

/* Creo Procedimiento para consultar contador en fuerte2 */
SET TERM !! ;
CREATE PROCEDURE sp_sfuerte2 (idclave numeric(3,0))
RETURNS (nultimo NUMERIC(2,0))
AS
BEGIN
   SELECT ultimo FROM fuerte2 WHERE clave = :idclave INTO :nultimo;   
END !!
SET TERM ; !!

/* Muestro las tablas creadas */
SHOW TABLE fuerte2;
SHOW TABLE fuerte;
SHOW TABLE debil;

/* Una vez que ha logrado insertar una fila en la tabla fuerte,
   debo replicar automaticamente dicha fila en la tabla fuerte2 */
SET TERM !! ;
CREATE TRIGGER trg_aifuerte FOR fuerte AFTER INSERT AS 
BEGIN
	/* INSERT INTO fuerte2 VALUES(NEW.clave,0); */
	EXECUTE PROCEDURE sp_idfuerte2 'I',NEW.clave;
END !!
SET TERM ; !!

/* Una vez que ha logrado borrar una fila en la tabla fuerte
   debo tambien borrar esa misma clave en la tabla fuerte2   */
SET TERM !! ;
CREATE TRIGGER trg_adfuerte FOR fuerte AFTER DELETE AS 
BEGIN
	/* DELETE FROM fuerte2 WHERE clave = OLD.clave; */
	EXECUTE PROCEDURE sp_idfuerte2 'D',OLD.clave;
END !!
SET TERM ; !!

/* Una vez que ha logrado hacer un cambio sobre el atributo 
   clave de la tabla fuerte, debo replicar el cambio sobre 
   la tabla fuerte2                                        */
SET TERM !! ;
CREATE TRIGGER trg_aufuerte FOR fuerte AFTER UPDATE AS 
BEGIN
   IF ( NEW.clave <> OLD.clave ) THEN 
	EXECUTE PROCEDURE sp_ucfuerte2 OLD.clave,NEW.clave;
  /*    UPDATE fuerte2 SET clave = NEW.clave WHERE clave = OLD.clave;*/
END !!
SET TERM ; !!

/* Creo Excepcion para manejar el posible error de discriminador
   no correlativo en tabla debil                           */
CREATE EXCEPTION ex_discr_nocorr 'Discriminador no correlativo y ascendente';

/* Antes de intentar insertar una tupla en la tabla debil, controlo 
   que el discriminador que pretende insertar sea igual al valor del
   ultimo insertado mas uno                                       */
SET TERM !! ;
CREATE TRIGGER trg_bidebil FOR debil BEFORE INSERT AS 
	DECLARE VARIABLE fultimo  NUMERIC(2,0);
BEGIN
/*        SELECT (ultimo+1) FROM fuerte2 WHERE clave = NEW.clave INTO :fultimo;*/
	EXECUTE PROCEDURE sp_sfuerte2 NEW.clave RETURNING_VALUES :fultimo;
	fultimo = fultimo + 1;
        IF ( :fultimo <> NEW.discr ) THEN EXCEPTION ex_discr_nocorr;
END !!
SET TERM ; !!

/* Una vez que ha logrado insertar una tupla en la tabla debil, sta se
   supone que es correlativa porque ya lo he controlado antes, en tal caso,
   solo me resta actualizar el ultimo valor de discriminador ingresado
   para esta clave                                                     */
SET TERM !! ;
CREATE TRIGGER trg_aidebil FOR debil AFTER INSERT AS 
BEGIN
/*        UPDATE fuerte2 SET ultimo = NEW.discr WHERE clave = NEW.clave;*/
	EXECUTE PROCEDURE sp_ufuerte2 NEW.clave,1;
END !!
SET TERM ; !!

/* Creo Excepcion para manejar el posible error que se genera cuando
   el usuario pretenda modificar algun valor de clave o discriminador
   en la tabla debil                                               */
CREATE EXCEPTION ex_discr_noupd 'Clave o Discriminador no actualizable';

/* Antes de intentar actualizar la clave de la tabla debil, genero el
   error correspondiente.                                          */
SET TERM !! ;
CREATE TRIGGER trg_budebil FOR debil BEFORE UPDATE AS 
BEGIN
        IF ( (OLD.clave <> NEW.clave) OR (OLD.discr <> NEW.discr) ) THEN 
           EXCEPTION ex_discr_noupd;
END !!
SET TERM ; !!

/* Creo Excepcion para manejar el posible error que se genera cuando
   el usuario pretende borrar una tupla que no posee el valor mas al-
   to de discriminador para la clave en cuestion.  Es posible borrar
   tuplas de la tabla debil solo "de mayor a menor"               */
CREATE EXCEPTION ex_discr_nodel 'Clave o Discriminador no borrable';

/* Antes de intentar borrar una tupla de la tabla debil, control que
   dicha tupla corresponda con el valor del ultimo discriminador in-
   gresado para dicha clave                                       */
SET TERM !! ;
CREATE TRIGGER trg_bddebil FOR debil BEFORE DELETE AS 
	DECLARE VARIABLE fultimo  NUMERIC(2,0);
BEGIN
/*        SELECT ultimo FROM fuerte2 WHERE clave = OLD.clave INTO :fultimo;*/
	EXECUTE PROCEDURE sp_sfuerte2 OLD.clave RETURNING_VALUES :fultimo;
        IF ( OLD.discr <> :fultimo ) THEN EXCEPTION ex_discr_nodel;
END !!
SET TERM ; !!

/* Una vez que ha logrado borrar una tupla de la tabla debil, se supone que
   dicha tupla poseia el valor de discriminador mas alto para la clave en
   cuestion, porque ya lo hemos controlado antes, en tal caso, debo decre-
   mentar el contador de ultimo en tabla fuerte2                        */
SET TERM !! ;
CREATE TRIGGER trg_addebil FOR debil AFTER DELETE AS 
BEGIN
/*        UPDATE fuerte2 SET ultimo = (ultimo - 1) WHERE clave = OLD.clave;*/
	EXECUTE PROCEDURE sp_ufuerte2 OLD.clave,-1;
END !!
SET TERM ; !!

/* Datos de Prueba - Datos buenos - Datos malos, pruebo stores, triggers */
INSERT INTO fuerte VALUES(1,'LIBRO 1');
INSERT INTO fuerte VALUES(2,'LIBRO 2');
INSERT INTO fuerte VALUES(3,'LIBRO 3');
INSERT INTO fuerte VALUES(4,'LIBRO 4');
COMMIT;
INSERT INTO debil  VALUES(1,1);
INSERT INTO debil  VALUES(1,2);
INSERT INTO debil  VALUES(1,5);
COMMIT;
UPDATE debil SET clave = 4;
UPDATE debil SET discr = 8;
/* UPDATE fuerte2 SET ultimo = 25; */
COMMIT;
DELETE FROM debil WHERE clave = 1 AND discr = 1;
COMMIT;
SELECT * FROM FUERTE2;
DELETE FROM debil WHERE clave = 1 AND discr = 2;
SELECT * FROM FUERTE2;
DELETE FROM debil WHERE clave = 1 AND discr = 1;
COMMIT;

SELECT * FROM FUERTE;

SELECT * FROM FUERTE2;

SELECT * FROM DEBIL;

DELETE FROM FUERTE WHERE clave = 4;
COMMIT;
SELECT * FROM fuerte;
SELECT * FROM fuerte2;

UPDATE fuerte SET clave = 18 WHERE clave = 1;
SELECT * FROM fuerte;
SELECT * FROM fuerte2;

