/*
Universidad Nacional de Lujn
Departamento de Ciencias Bsicas
Division Estadstica y Sistemas
Base de Datos
Lic. Guillermo Cherencio

Implementacion de Generalizacion Obligatoria Exclusiva
con FireBird / InterBase

La tabla padre es la generica y las tablas hijo1, hijo2 representan
lo especifico.  La obligatoriedad de padre con respecto a hijo1, hi-
jo2 debera realizarse a traves de la ejecucion de una transaccion en
donde se pretenda insertar en la tabla padre y en la tabla hijo1 o 
bien en la tabla hijo2 (pero no en ambas) al mismo tiempo.  En cuan-
to a la exclusividad la podemos controlar a traves de triggers sobre
las tablas hijo1 e hijo2 en donde se controlara sobre la tabla opues-
ta la no existencia de la clave que se pretende insertar o cambiar y
viceversa.

Modo de Ejecucion:  	
   isql ferreteria.gdb -u SYSDBA -p masterkey -i dersql2.sql -o salida.txt

*/

SET ECHO ON;
SET;

DROP TRIGGER trg_bihijo2;
DROP TRIGGER trg_bihijo1;
DROP TRIGGER trg_buhijo2;
DROP TRIGGER trg_buhijo1;
DROP PROCEDURE sp_shijo12;
DROP EXCEPTION ex_esta_hijo2;
DROP EXCEPTION ex_esta_hijo1;
DROP TABLE hijo1;
DROP TABLE hijo2;
DROP TABLE padre;

/* Creo Tablas */
CREATE TABLE padre(
   clave NUMERIC(3,0) DEFAULT 1 NOT NULL,
   descr VARCHAR(40) NOT NULL,
   PRIMARY KEY (clave)
);

CREATE TABLE hijo1(
   clave NUMERIC(3,0) DEFAULT 1 NOT NULL REFERENCES padre,  /* fk entidad padre */
   athijo1 NUMERIC(2,0) DEFAULT 1 NOT NULL,  /* atributo propio de hijo1 */
   PRIMARY KEY (clave)
);

CREATE TABLE hijo2(
   clave NUMERIC(3,0) DEFAULT 1 NOT NULL REFERENCES padre,  /* fk entidad padre */
   athijo2 NUMERIC(2,0) DEFAULT 1 NOT NULL,  /* atributo propio de hijo2 */
   PRIMARY KEY (clave)
);

SHOW TABLE padre;
SHOW TABLE hijo1;
SHOW TABLE hijo2;

/* Creo Procedimiento para consultar tabla hijo1 o hijo2 */
SET TERM !! ;
CREATE PROCEDURE sp_shijo12 (tabla char(1),idclave numeric(3,0))
RETURNS (ntuplas NUMERIC(2,0))
AS
BEGIN
   IF ( :tabla = '1' ) THEN SELECT COUNT(*) FROM hijo1 WHERE clave = :idclave INTO :ntuplas;   
   IF ( :tabla = '2' ) THEN SELECT COUNT(*) FROM hijo2 WHERE clave = :idclave INTO :ntuplas;   
END !!
SET TERM ; !!

/* Se puede ingresar en hijo1 si no esta en hijo2 */
CREATE EXCEPTION ex_esta_hijo2 'Esta clave ya existe en hijo2';

SET TERM !! ;
CREATE TRIGGER trg_bihijo1 FOR hijo1 BEFORE INSERT AS 
	DECLARE VARIABLE esta SMALLINT;
BEGIN
	EXECUTE PROCEDURE sp_shijo12 '2',NEW.clave RETURNING_VALUES :esta;
/*        SELECT COUNT(*) FROM hijo2 WHERE clave = NEW.clave INTO :esta;*/
        IF ( :esta > 0 ) THEN EXCEPTION ex_esta_hijo2;
END !!
SET TERM ; !!

SET TERM !! ;
CREATE TRIGGER trg_buhijo1 FOR hijo1 BEFORE UPDATE AS 
	DECLARE VARIABLE esta SMALLINT;
BEGIN
	EXECUTE PROCEDURE sp_shijo12 '2',NEW.clave RETURNING_VALUES :esta;
/*        SELECT COUNT(*) FROM hijo2 WHERE clave = NEW.clave INTO :esta;*/
        IF ( :esta > 0 ) THEN EXCEPTION ex_esta_hijo2;
END !!
SET TERM ; !!

/* Se puede ingresar en hijo2 si no esta en hijo1 */
CREATE EXCEPTION ex_esta_hijo1 'Esta clave ya existe en hijo1';

SET TERM !! ;
CREATE TRIGGER trg_bihijo2 FOR hijo2 BEFORE INSERT AS 
	DECLARE VARIABLE esta SMALLINT;
BEGIN
	EXECUTE PROCEDURE sp_shijo12 '1',NEW.clave RETURNING_VALUES :esta;
/*        SELECT COUNT(*) FROM hijo1 WHERE clave = NEW.clave INTO :esta;*/
        IF ( :esta > 0 ) THEN EXCEPTION ex_esta_hijo1;
END !!
SET TERM ; !!

SET TERM !! ;
CREATE TRIGGER trg_buhijo2 FOR hijo2 BEFORE UPDATE AS 
	DECLARE VARIABLE esta SMALLINT;
BEGIN
	EXECUTE PROCEDURE sp_shijo12 '1',NEW.clave RETURNING_VALUES :esta;
/*        SELECT COUNT(*) FROM hijo1 WHERE clave = NEW.clave INTO :esta;*/
        IF ( :esta > 0 ) THEN EXCEPTION ex_esta_hijo1;
END !!
SET TERM ; !!

/* Datos de Prueba */
INSERT INTO padre VALUES(1,'PADRE 1');
INSERT INTO padre VALUES(2,'PADRE 2');
INSERT INTO padre VALUES(3,'PADRE 3');
INSERT INTO padre VALUES(4,'PADRE 4');
COMMIT;
SELECT * FROM padre;
INSERT INTO hijo1  VALUES(1,1);
INSERT INTO hijo1  VALUES(2,1);
INSERT INTO hijo1  VALUES(3,1);
COMMIT;
INSERT INTO hijo2  VALUES(4,1);
COMMIT;
SELECT * FROM hijo1;
SELECT * FROM hijo2;
INSERT INTO hijo2 VALUES(3,1);
SELECT * FROM hijo2;
INSERT INTO hijo1 VALUES(4,1);
SELECT * FROM hijo1;
COMMIT;

UPDATE hijo2 SET clave = 1 WHERE clave = 4;
SELECT * FROM hijo2;
UPDATE hijo2 SET athijo2 = 44;
SELECT * FROM hijo2;
COMMIT;

